/*
 * Copyright 2008- kensir0u.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
/*
 * 쐬       F2008/11/05
 * pbP[W   Fikdnet.dijava
 * t@C   FTokenizer.java
 */
package ikdnet.dijava;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * <i>Tv</i>F g[NNX @.
 * <p>
 * g[NNX
 * 
 * 
 * @author kensir0u
 * @version 1.0
 * @since JDK 5.0
 * 
 */
public class Tokenizer {

	/**
	 * 
	 * <i>Tv</i>F t@Ĉ̕ԋpB
	 * .<p>
	 * t@Ĉ̕ԋpsBg[NXguclassvg[N̎̃V[PX
	 * o[ԋpB
	 *@
	 * @param list
	 * @return@t@ĆiNX́j
	 */
	public static String getFileName(final List<Token> list) {

		for (int i = 0; i < list.size(); i++) {
			Token t = list.get(i);
			if ("class".equals(t.getSentence())) {
				if (i + 1 < list.size()) {
					t = list.get(i + 1);
					return t.getSentence();
				} else {
					break;
				}
			}
		}
		return null;
	}

	/**
	 * 
	 * <i>Tv</i>F lXg̃JEgsB
	 * .<p>
	 * ʂ̃yA`FbN̂߂̃lXgJEgAlԋpB
	 *@yAƃ[ԋpB@
	 *
	 * @param list
	 * @return@lXgJEg@
	 */
	public static int getNestCount(final List<Token> list) {
		int nest = 0;
		for (int i = 0; i < list.size(); i++) {
			final Token t = list.get(i);
			if ("{".equals(t.getSentence())) {
				nest++;
			} else if ("}".equals(t.getSentence())) {
				nest--;
			}
		}
		return nest;
	}

	/**
	 * \@ꗗ@
	 */
	private static final String[] reservedWords = { "++", "+=", "--", "-=",
			"**", "*=", "/=", "==", "!=", "&&", "||", "&=", "|=", "abstract",
			"boolean", "break", "byte", "case", "catch", "char", "class",
			"continue", "default", "do", "else", "extends", "false", "final",
			"float", "for", "implements", "import", "int", "interface", "new",
			"null", "package", "private", "protected", "public", "return",
			"short", "static", "super", "switch", "synchronized", "this",
			"throws", "try", "true", "void", "while", };

	private static char[] buf;
	private static int index;
	private static int lastIndex;
	private static Map<String,String> reservedWord = new HashMap<String,String>();

	public static void main(final String args[]) {

		final String source = "package just.generated;\n"
				+ "import java.lang.annotation.Retention; //eXgRg\n"
				+ "import java.lang.annotation.RetentionPolicy;\n"
				+ "public class Hello {\n"
				+ "@Retention(RetentionPolicy.RUNTIME)\n"
				+ "@interface StringAnnotation {\n" + "String value();\n"
				+ "}\n" + "   @StringAnnotation(\"b\h\")\n"
				+ "	public static void main(String... args) {\n"
				+ "		System.out.println(new Object() {\n"
				+ "			public String toString() {\n"
				+ "				return \"just hellofasdf\\\"das!\";\n" + "			}\n"
				+ "		});\n" + "	}\n" + "}\n";

		final List<Token> l = split(source);

		for (final Token s : l) {
			System.out.println(s.getSentence());
		}
	}
	
	/**
	 * 
	 * <i>Tv</i>F \[X̃g[N
	 * .<p>
	 * \[X̃g[NsB
	 *@
	 * @param source
	 * @return@g[NXg
	 */
	public static List<Token> split(final String source) {

		buf = source.toCharArray();
		index = 0;
		lastIndex = buf.length;

		if (reservedWord.size() == 0) {
			for (int i = 0; i < reservedWords.length; i++) {
				reservedWord.put(reservedWords[i], reservedWords[i]);
			}
		}
		final List<Token> list = new ArrayList<Token>();

		while (isGetChar()) {

			final char c = getChar();

			if (Character.isWhitespace(c)) {
				continue;
			}

			if (Character.isDigit(c)) { // l
				list.add(readNumber(c));
				continue;
			}

			if (c == '/') {
				list.add(readComment(c));
				continue;
			}

			if (c == '"') {
				list.add(readString()); // 
			} else if (c == '+' || c == '-' || c == '*' || c == '/' || c == '|'
					|| c == '&' || c == '!' || c == '>' || c == '<') {
				list.add(readSymbol(c));

			} else if (isSymbol(c)) {
				list.add(new Token("" + c, "" + c, TokenType.BRACE));// L
			} else {
				list.add(readIdentifier(c));
			} // O

		}

		return list;
	}

	/**
	 * 
	 * <i>Tv</i>F Lǂ𔻒fB
	 * .<p>
	 * Lǂ̔fsB
	 *@
	 * @param ch
	 * @return@Lf
	 */
	private static boolean isSymbol(final char ch) {
		if (ch < 'A')
			return true;
		if (ch == '{' || ch == '}' || ch == '[' || ch == ']')
			return true;
		return false;
	}

	/**
	 * 
	 * <i>Tv</i>F LN^[f[^̎擾
	 * .<p>
	 * LN^[f[^̎擾sB
	 *@
	 * @return@Lf[^
	 */
	private static char getChar() {
		final char c = buf[index];
		index++;
		return c;
	}

	/**
	 * 
	 * <i>Tv</i>F LN^[f[^擾\fB
	 * .<p>
	 * LN^[f[^擾\fB
	 *@
	 * @return@f
	 */
	private static boolean isGetChar() {
		return index < lastIndex;
	}

	/**
	 * 
	 * <i>Tv</i>F LN^[f[^擾
	 * .<p>
	 * ̃LN^[f[^擾B
	 *@
	 * @return@Lf[^
	 */
	private static char nextChar() {
		return buf[index];
	}

	/**
	 * 
	 * <i>Tv</i>FLN^[f[^擾\fB
	 * .<p>
	 * ̃LN^[f[^擾\fB
	 *@
	 * @return@f
	 */
	private static boolean isNextChar() {
		return (index) < lastIndex;
	}

	/**
	 * 
	 * <i>Tv</i>F lf[^̓ǂݍ
	 * .<p>
	 * lf[^̓ǂݍ
	 *@
	 * @param head
	 * @return@lg[N
	 */
	private static Token readNumber(final char head) {
		final StringBuffer str = new StringBuffer().append(head);
		while (isNextChar()) {
			if (!Character.isDigit(nextChar())) {
				break;
			}
			str.append(getChar());

		}
		return new Token(str.toString(), str.toString(), TokenType.NUMBER,
				Long.class);
	}

	/**
	 * 
	 * <i>Tv</i>F L̓ǂݍ
	 * .<p>
	 * L̓ǂݍ
	 *@
	 * @param c
	 * @return@Lg[N
	 */
	private static Token readSymbol(final char c) {
		final char nc = nextChar();
		if (c == nc || nc == '=') { // ++, +=, --, -=, **, *=, //, /=,
			// &&, &=, ||, |=, !!, !=, >>, <<
			getChar();
			return new Token("" + c + nc, "" + c + nc, TokenType.OPE,
					String.class);
		}
		return new Token(String.valueOf(c), String.valueOf(c), TokenType.OPE,
				String.class);
	}

	/**
	 * 
	 * <i>Tv</i>F ʎq̓ǂݍ
	 * .<p>
	 * ʎq̓ǂݍ
	 *@
	 * @param head
	 * @return@ʎqg[N
	 */
	private static Token readIdentifier(final char head) {
		final StringBuffer str = new StringBuffer().append(head);
		while (isNextChar()) {
			if (!Character.isLetterOrDigit(nextChar())) {
				break;
			}
			str.append(getChar());

		}

		final String word = str.toString();
		// \ꂩǂH
		if (reservedWord.containsKey(word)) {
			return new Token(word, word, TokenType.KEYWORD, String.class);
		} else {
			return new Token(word, word, TokenType.IDENTIFIER, String.class);
		}
	}

	/**
	 * 
	 * <i>Tv</i>F ̓ǂݍ
	 * .<p>
	 * ̓ǂݍ
	 *@
	 * @return@g[N
	 */
	private static Token readString() {
		final StringBuffer str = new StringBuffer();
		while (isGetChar()) {
			final char c = getChar();
			str.append(c);
			if (!isNextChar())
				break;
			if ('"' == nextChar() && '\\' != c) {
				getChar();
				break;
			}

		}
		return new Token(str.toString(), str.toString(), TokenType.STRING,
				String.class);
	}

	/**
	 * 
	 * <i>Tv</i>F Rg̓ǂݍ
	 * .<p>
	 * Rg̓ǂݍ݂sB
	 *@isRgAsRgj
	 * @param c
	 * @return@Rgg[N
	 */
	private static Token readComment(final char c) {

		final char nch = nextChar();
		if (nch == '/') {
			return readLF(c);
		} else if (nch == '*') {
			return readMultiComment(c);
		} else {
			return new Token("", String.valueOf(c) + String.valueOf(nch),
					TokenType.ERROR, String.class);
		}
	}

	/**
	 * 
	 * <i>Tv</i>F PsRg̓ǂݍ
	 * .<p>
	 * Ps̓ǂݍ
	 *@
	 * @param head
	 * @return@PsRgg[N
	 */
	private static Token readLF(final char head) {
		final StringBuffer str = new StringBuffer();
		str.append(head);
		char c = getChar();
		str.append(c);
		while (c != '\n' && c != '\0') {
			c = getChar();
			str.append(c);
		}
		return new Token(str.toString(), str.toString(), TokenType.COMMENT,
				String.class);
	}

	/**
	 * 
	 * <i>Tv</i>F sRg̓ǂݍ
	 * .<p>
	 * sRg̓ǂݍ
	 *@
	 * @param head
	 * @return@sRgg[N
	 */
	private static Token readMultiComment(final char head) {
		final StringBuffer str = new StringBuffer();
		str.append(head);
		char c = getChar();
		str.append(c);
		while (isGetChar()) {
			c = getChar();
			str.append(c);
			final char nch = nextChar();
			if (c == '*' && nch == '/') {
				str.append(getChar());
				break;
			}
		}
		return new Token(str.toString(), str.toString(), TokenType.COMMENT,
				String.class);
	}

}
